@@ -1,16 +1,16 @@
 # frozen_string_literal: true
 
 module ActiveModel
-  # Returns the version of the currently loaded \Active \Model as a <tt>Gem::Version</tt>
+  # Returns the currently loaded version of \Active \Model as a <tt>Gem::Version</tt>.
   def self.gem_version
     Gem::Version.new VERSION::STRING
   end
 
   module VERSION
-    MAJOR = 6
+    MAJOR = 7
     MINOR = 0
-    TINY  = 6
-    PRE   = "1"
+    TINY  = 4
+    PRE   = nil
 
     STRING = [MAJOR, MINOR, TINY, PRE].compact.join(".")
   end
